//! Fragment management
//!
//! Re-exports fragment types from types.rs.
//! See [`crate::vector::types::Fragment`] for the main fragment type.

pub use crate::vector::types::{Fragment, FragmentState};
